/* ********************************************
 *  
 *  Battery Powered Mini Scene Controller
 *  
 *  This is a scene controller using the MySensors library and
 *  has a small OLED display to provide the labels for the buttons
 *  and status update:
 *     o Scene Started/Stopped
 *     o Time
 *     o Date
 *     o Temperature
 *     
 *  It also used a cut down version of the Adafruit SSD1306/GFX libraries
 *  to control the OLED display. The library had to be cut doen to
 *  make it fit along side the MySensors library.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  version 3 as published by the Free Software Foundation.
 *  
 * ********************************************/

#include  <MySensor.h>
#include  <SPI.h>
#include  <Wire.h>
#include  <Time.h>
#include  <stdarg.h>
#include  <SSD1306_CHAR_GFX.h>
#include  <SSD1306_CHAR.h>
#include  <Bounce2.h>

#define   SKETCH_NAME "Mini Ctrl"
#define   SKETCH_MAJOR_VER  "0"
#define   SKETCH_MINOR_VER  "3"
#define   CHILD_ID 1
#define   REFRESH_TIME 43200000 // wake every 12 hours to update
#define   DISPLAY_TIME 30000    // keep screen active for 30 seconds
#define   ASCII_BASE_ONLY       // only allow base printable ASCII chars to same memory
#define   CONFIG_RATE 3000
#define   MESSAGE_DELAY 1500
#define   BUTTON_INT_PIN 3
#define   OLED_RESET -1

boolean   WAKE_BY_PIN = true;
byte      BUTTON_PINS[]  = {4,5,6,7};
bool      SCENE_STATES[] = {0,0,0,0};
char *    BUTTON_TEXT[] = {"Btn01", "Btn02", "Btn03", "Btn04"};
char      TEMP_STRING[] = "          ";
unsigned long WAKE_TIME = millis();
unsigned long LAST_CONFIG = millis();
unsigned long MESSAGE_TIME;
boolean   MESSAGE_DISPLAY = false;
boolean   DO_CONFIG = true;
byte      CONFIG_ITEM = 0;
boolean   BUTTON_RELEASE = true;


SSD1306_CHAR   display(OLED_RESET);
MyMessage msg_on(CHILD_ID, V_SCENE_ON);
MyMessage msg_off(CHILD_ID, V_SCENE_OFF);
MySensor  sensor_node;
Bounce    debouncer = Bounce();

void setup()  
{  
  Serial.begin(112500);
  debouncer.attach(BUTTON_INT_PIN);
  debouncer.interval(10);

  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);  // initialize with the I2C addr 0x3C (for the 128x64)
  sensor_node.begin(incomingMessage);

  // Send the sketch version information to the gateway and Controller
  sensor_node.sendSketchInfo(SKETCH_NAME, SKETCH_MAJOR_VER"."SKETCH_MINOR_VER);

  // Setup the interrupt pin
  pinMode(BUTTON_INT_PIN, INPUT);
  digitalWrite(BUTTON_INT_PIN, HIGH);

  // Present scene controller to controller
  sensor_node.present(CHILD_ID, S_SCENE_CONTROLLER);
  sensor_node.request(CHILD_ID,V_TEMP);
  sensor_node.requestTime(receiveTime); 

  // Setup the buttons
  byte i;
  for (i=0;i<4;i++) {
    pinMode(BUTTON_PINS[i], INPUT);
    digitalWrite(BUTTON_PINS[i], HIGH);
  }

  startConfig();

  // set text style
  display.clearDisplay();
  displayButtons();
//  clearCentrePanel();
//  displayCentrePanel();
}


void loop() 
{

  uint8_t value;
  uint8_t i;
  boolean buttonChanged;

  doConfig();            // check for configuration changes
  sensor_node.process(); // process any inbound messages

  buttonChanged = debouncer.update();

  if (  buttonChanged && !debouncer.read() ) { // button pushed active low
      WAKE_TIME = millis();                    // update wake time for new activity
      sensor_node.request(CHILD_ID,V_TEMP);
      sensor_node.requestTime(receiveTime); 

      for (i=0; i<4; i++ ) checkButton(i);     // find out which button has changed state
  }

  // Check to see if we need to remove the message
  if ( MESSAGE_DISPLAY && timeDiff( MESSAGE_TIME, MESSAGE_DELAY )) {
    MESSAGE_DISPLAY = false;
    displayMiddle();
  }

  // sleep
  if ( timeDiff( WAKE_TIME, DISPLAY_TIME )) {  // need to put check here so that we don't sleep when just asked for the config data.

    // Sleep until something happens
    display.ssd1306_command(SSD1306_DISPLAYOFF);
    WAKE_BY_PIN = sensor_node.sleep(BUTTON_INT_PIN - 2, FALLING, REFRESH_TIME);

    // Just woke up
    WAKE_TIME = millis();
    LAST_CONFIG = millis();

    // Turn on screen
    if ( WAKE_BY_PIN ) {
      sensor_node.request(CHILD_ID,V_TEMP);
      display.ssd1306_command(SSD1306_DISPLAYON);
      sensor_node.requestTime(receiveTime); 
    }
    
    // Refresh config
    startConfig();
  }
} 

void startConfig() {
  if (!DO_CONFIG ) CONFIG_ITEM = 0;
  DO_CONFIG = true;
}

void sendBattery( void ) {
  uint16_t milliVolts = hwCPUVoltage();
  byte pc = milliVolts/30;
  sensor_node.sendBatteryLevel( (pc>100?100:pc) );
}

void doConfig() {  // slow down request traffic to prevent network flooding
  if ( timeDiff( LAST_CONFIG, CONFIG_RATE )) {
    LAST_CONFIG = millis();
    switch ( CONFIG_ITEM ) {
      case  0:  sendBattery(); break;
      case  1:  sensor_node.request(CHILD_ID,V_VAR1); break;
      case  2:  sensor_node.request(CHILD_ID,V_VAR2); break;
      case  3:  sensor_node.request(CHILD_ID,V_VAR3); break;
      case  4:  sensor_node.request(CHILD_ID,V_VAR4); break;
      default:  DO_CONFIG = false;  // config is complete
    }
    CONFIG_ITEM++;
  }
}

uint16_t hwCPUVoltage() {
  // Code copied from within the MySensors libraries
  // Measure Vcc against 1.1V Vref 
  #if defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
    ADMUX = (_BV(REFS0) | _BV(MUX4) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1));
  #elif defined (__AVR_ATtiny24__) || defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
    ADMUX = (_BV(MUX5) | _BV(MUX0));
  #elif defined (__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__)
    ADMUX = (_BV(MUX3) | _BV(MUX2));
  #else
    ADMUX = (_BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1));
  #endif 
  // Vref settle
  delay(70); 
  // Do conversion
  ADCSRA |= _BV(ADSC);
  while (bit_is_set(ADCSRA,ADSC));
  // return Vcc in mV
  return (1125300UL) / ADC;
}


void checkButton( int id ) {
  bool value;

  if (!digitalRead(BUTTON_PINS[id]) ) { // button pressed active low
     // button has been pressed
     SCENE_STATES[id] = !SCENE_STATES[id];
     Serial.println(SCENE_STATES[id]==HIGH ? " Started" : " Stopped");

     if ( SCENE_STATES[id] ) {
       sensor_node.send(msg_on.set(id));
       printMessage("Started");
     } else {
       sensor_node.send(msg_off.set(id));
       printMessage("Stopped");
     }
  }
}

void incomingMessage(const MyMessage &message) {
  
  if (message.isAck()) {
     // do nothing
  }
  
  switch ( message.type ) {
    case V_VAR1: setButtonText( 0, message.getString(), mGetLength(message) ); break;
    case V_VAR2: setButtonText( 1, message.getString(), mGetLength(message) ); break;
    case V_VAR3: setButtonText( 2, message.getString(), mGetLength(message) ); break;
    case V_VAR4: setButtonText( 3, message.getString(), mGetLength(message) ); break;
    case V_TEMP: setTemp(message.getString(), mGetLength(message) ); break;
  }
}

bool timeDiff( unsigned long lastUpdate, unsigned long interval ) {
  unsigned long currentTime = millis();
  unsigned long timeDiff = 0UL;
  
    if ( currentTime >= lastUpdate ) {
      timeDiff = currentTime - lastUpdate;
    } else { 
      timeDiff = 0xFFFFFFFFUL - lastUpdate + currentTime;
    }
    
    if ( timeDiff >= interval ) {
      return true;
    } else {  
      return false;
    }
}

void displayButtons() {
  displayButton(0);
  displayButton(1);
  displayButton(2);
  displayButton(3);
  drawLines();
}


void displayButton(uint8_t button) {

  display.setTextWrap(0);
  display.setTextSize(2);
//  display.setTextColor(WHITE, BLACK);

  switch (button) {
  // left buttons
  case 0:
    display.setCursor(0,0);
    display.print(BUTTON_TEXT[0]);
    break;
    
  case 1:
    display.setCursor(0,6);
    display.print(BUTTON_TEXT[1]);
    break;
  
  // right buttons
  case 2:
    display.setCursor(16-(strlen(BUTTON_TEXT[2])),0);
    display.print(BUTTON_TEXT[2]);
    break;
  
  case 3:
    display.setCursor(16-(strlen(BUTTON_TEXT[3])),6);
    display.print(BUTTON_TEXT[3]);
    break;
  }
}

void receiveTime(unsigned long time) {
  setTime(time);
  if ( !MESSAGE_DISPLAY ) {
    displayTime();
    displayDate();
  }
}

void setButtonText(byte button, const char * text, byte len) {
  byte j;
  byte _len = (len>5)?5:len;
  byte offset = (button==2||button==3)?5-_len:0;                  // offset for left or right justification

  for (j=0; j<5; j++) BUTTON_TEXT[button][j]=' ';                 // Blank string
  strncpy(BUTTON_TEXT[button]+offset, text, _len);
  BUTTON_TEXT[button][5]=0x00;                                    // NULL terminate the string

  displayButton(button);
}

void displayTime() {
  display.setTextSize(2);
  display.setCursor(0,3);
  
  if ( hour() < 10 ) display.print("0");
  display.print(hour());
  display.print(":");
  if ( minute() < 10 ) display.print("0");
  display.print(minute());
}

void displayDate() {
  display.setTextWrap(0);
  display.setTextSize(1);
  display.setCursor(11,4);
  display.print(dayShortStr(weekday()));
  display.print(" ");
  display.print(day());
  display.print(" ");
  display.print(monthShortStr(month())); 
}


void setTemp(const char * temp, byte len ) {
  byte j;
  byte _len = (len>5)?5:len;

  for (j=0;j<11;j++) TEMP_STRING[j] = ' ';
  strncpy(TEMP_STRING+(9-_len), temp, _len);
  TEMP_STRING[9] = 'C';
  TEMP_STRING[10] = 0x00;                              // NULL terminate the string
  if ( !MESSAGE_DISPLAY ) displayTemp();
}

void displayTemp() {
  display.setTextWrap(0);
  display.setTextSize(1);
  display.setCursor(11,3);
  display.print(TEMP_STRING);                         // display the temp
}

void eraseMiddle() {
  display.setTextSize(2);
  display.setCursor(0,3);
  for (int i=0; i<11; i++) display.print(" ");
}

void displayMiddle() {
  eraseMiddle();
  displayTime();
  displayTemp();
  displayDate();  
}

void printMessage( char * msg ) {
  MESSAGE_TIME = millis();
  MESSAGE_DISPLAY = true;
  eraseMiddle();
  display.setTextSize(2);
  display.setCursor(0,3);
  for ( int i=0; i<5-(strlen(msg)/2); i++) display.print(" ");
  display.print( msg );
}

void drawLines() {
  uint8_t TOPLINE[] = {2,2,2,2,2,2};
  uint8_t BOTTOMLINE[] = {64,64,64,64,64,64};
  uint8_t SIDELINES[] = {0,255,0,0,255,0};
  uint8_t TOPCORNERS[] = {2,3,0,0,3,2};
  uint8_t BOTTOMCORNERS[] = {64,192,0,0,192,64};
  
  display.setTextSize(1);

  // Top
  for ( int i=0; i<10; i++ ) display.drawBlock(i,2,TOPLINE,6);
  for ( int i=11; i<21; i++ ) display.drawBlock(i,2,TOPLINE,6);
  display.drawBlock(10,0,SIDELINES,6); 
  display.drawBlock(10,1,SIDELINES,6);
  display.drawBlock(10,2,TOPCORNERS,6);

  // Bottom
  for ( int i=0; i<10; i++ ) display.drawBlock(i,5,BOTTOMLINE,6);
  for ( int i=11; i<21; i++ ) display.drawBlock(i,5,BOTTOMLINE,6);
  display.drawBlock(10,6,SIDELINES,6); 
  display.drawBlock(10,7,SIDELINES,6);
  display.drawBlock(10,5,BOTTOMCORNERS,6);
}

